import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize
from scipy.stats import multivariate_normal as mvn

from .utils import trans_UC, BIC, AIC, LME

#===============================================================================
#
#   FITRMODEL
#       The highest level object representing a task model to be fit.
#       It is not necessary to use this object, since one could use the EM
#        object (lower level), but this method should be the simplest for
#        inexperienced users
#
#===============================================================================

class fitrmodel(object):
    """
    An object representing a model to be fit to behavioural data. This should be viewed as a high level wrapper for multiple potential model fitting algorithms which themselves can be run by using their respective classes.

    Attributes
    ----------
    name : str
        Name of the model. We suggest identifying model based on free parameters.
    loglik_func : function
        The log-likelihood function to be used to fit the data
    params : list
        List of reinforcement learning parameter objects from the rlparams module.
    """
    def __init__(self, loglik_func, params, name=None):
        self.name = name
        self.loglik_func = loglik_func
        self.params = params

    def fit(self, data, method='EM', c_limit=0.01):
        """
        Runs model fitting

        Parameters
        ----------
        method : {'EM', 'MLE'}
            The inference algorithm to use.
        c_limit : float
            Limit at which convergence of log-posterior probability is determined

        Returns
        -------
        fitrfit : object
            Representation of the model fitting results
        """

        if method=='EM':
            opt = EM(loglik_func=self.loglik_func, params=self.params, name=self.name)
            results = opt.fit(data=data, c_limit=c_limit)
        elif method=='MLE':
            opt = MLE(loglik_func=self.loglik_func, params=self.params, name=self.name)
            results = opt.fit(data=data, c_limit=c_limit)

        return results

#===============================================================================
#
#   INFERENCE METHOD OBJECTS
#       - EM: Expectation-Maximization with Laplace Approximation
#       - MLE: Maximum Likelihood Estimate
#
#===============================================================================

class EM(object):
    """
    Expectation-Maximization with the Laplace Approximation

    Attributes
    ----------
    name : str
        Name of the model being fit. We suggest using the free parameters.
    loglik_func : function
        The log-likelihood function to be used for model fitting
    params : list
        List of parameters from the rlparams module
    nparams : int
        Number of free parameters in the model
    param_rng : list
        List of strings denoting the parameter ranges (see rlparams module for further details)
    prior : scipy.stats distribution
        The prior distribution over parameter estimates. Here this is fixed to a multivariate normal.
    mu : ndarray(shape=nparams)
        The prior mean over parameters
    cov : ndarray(shape=(nparams,nparams))
        The covariance matrix for prior over parameter estimates

    References
    ----------
    [1] Huys, Q. J. M., et al. (2011). Disentangling the roles of approach, activation and valence in instrumental and pavlovian responding. PLoS Computational Biology, 7(4).
    """
    def __init__(self, loglik_func, params, name=None):
        self.name = name
        self.loglik_func = loglik_func
        self.params = params

        # Extract properties of parameters that will be used in later functions
        self.nparams = len(params)
        self.param_rng = []
        for i in range(self.nparams):
            self.param_rng.append(params[i].rng)


        # Initialize prior
        self.prior = mvn
        self.mu  = np.zeros(self.nparams)
        self.cov = np.eye(self.nparams)

    def fit(self, data, n_iterations=1000, c_limit=1, opt_algorithm='BFGS'):
        """
        Performs maximum a posteriori estimation of subject-level parameters

        Parameters
        ----------
        data : dict
            Dictionary of data from all subjects.
        n_iterations : int
            Maximum number of iterations to allow.
        c_limit : float
            Threshold at which convergence is determined
        opt_algorithm : {'BFGS', 'L-BFGS-B', 'Nelder-Mead'}
            Algorithm to use for optimization

        Returns
        -------
        fitrfit : object
            Representation of the model fitting results
        """

        # Instantiate the results object
        nsubjects = len(data)
        results = fitrfit(method='Expectation-Maximization',
                          nsubjects=nsubjects,
                          nparams=self.nparams,
                          name=self.name)
        results.set_paramnames(params=self.params)

        convergence = False
        opt_iter = 1
        sum_nlogpost = 0 # Monitor total neg-log-posterior for convergence
        while convergence == False and opt_iter < n_iterations:
            for i in range(nsubjects):
                print('ITERATION: '          + str(opt_iter) +
                      ' | [E-STEP] SUBJECT: ' + str(i+1))

                # Extract subject-level data
                S = data[i]['S']
                A = data[i]['A']
                R = data[i]['R']

                # Construct the subject's negative log-posterior function
                _nlogpost = lambda x: -self.logposterior(x=x, states=S, actions=A, rewards=R)

                # Generate initial values
                x0 = np.random.normal(loc=0, scale=2, size=self.nparams)
                # Optimize
                res = minimize(_nlogpost, x0, method=opt_algorithm)

                # Update subject level data if logposterior improved
                if res.fun < results.nlogpost[i]:
                    results.nlogpost[i] = res.fun
                    results.params[i,:] = res.x
                    results.hess[:,:,i] = np.linalg.inv(res.hess_inv)
                    results.hess_inv[:,:,i] = res.hess_inv
                    results.nloglik[i]  = self.loglik_func(params=trans_UC(res.x, rng=self.param_rng), states=S, actions=A, rewards=R)
                    results.LME[i] = LME(-res.fun, self.nparams, np.linalg.inv(res.hess_inv))
                    results.AIC[i] = AIC(self.nparams, results.nloglik[i])
                    results.BIC[i] = BIC(results.nloglik[i], self.nparams, len(data[0]['A']))

            # If the group level posterior probability has converged, stop
            if np.abs(np.sum(results.nlogpost) - sum_nlogpost) < c_limit:
                convergence = True
            else:
                # Update the running model log-posterior probability
                sum_nlogpost = np.sum(results.nlogpost)

                # Add total LME, BIC, and AIC to results list
                results.ts_LME.append(np.sum(results.LME))
                results.ts_BIC.append(np.sum(results.BIC))
                results.ts_AIC.append(np.sum(results.AIC))

                # Optimize group level hyperparameters
                print('\nITERATION: ' + str(opt_iter) + ' | [M-STEP]\n')
                self.group_level_estimate(param_est=results.params,
                                          hess_inv=results.hess_inv)
                opt_iter += 1

        # Transform data to constrained space
        for i in range(nsubjects):
            results.params[i,:] = trans_UC(results.params[i,:], self.param_rng)

        return results

    def logposterior(self, x, states, actions, rewards):
        """
        Represents the log-posterior probability function

        Parameters
        ----------
        x : ndarray(nparams)
            Array of parameters for single subject
        states : ndarray
            Array of states encountered by subject. Number of rows should reflect number of trials. If the task is a multi-step per trial task, then the number of columns should reflect the number of steps, unless a custom likelihood function is used which does not require this.
        actions: ndarray
            Array of actions taken by subject. Number of rows should reflect number of trials. If the task is a multi-step per trial task, then the number of columns should reflect the number of steps, unless a custom likelihood function is used which does not require this.
        rewards : ndarray
            Array of rewards received by the subject. Number of rows should reflect number of trials. If there are multiple steps at which rewards are received, they should be stored in different columns, unless a custom likelihood funciton is used which does not require this.

        Returns
        -------
        float
            Log-posterior probability
        """

        lp = self.loglik_func(params=trans_UC(x, rng=self.param_rng), states=states, actions=actions, rewards=rewards) + self.prior.logpdf(x, mean=self.mu, cov=self.cov)

        return lp

    def group_level_estimate(self, param_est, hess_inv):
        """
        Updates the group-level hyperparameters

        Parameters
        ----------
        param_est : ndarray(shape=(nsubjects, nparams))
            Current parameter estimates for each subject
        hess_inv : ndarray(shape=(nparams, nparams, nsubjects))
            Inverse Hessian matrix estimate for each subject from the iteration with highest log-posterior probability
        """
        nsubjects = np.shape(param_est)[0]

        self.mu = np.mean(param_est, axis=0)
        self.cov = np.zeros([self.nparams, self.nparams])

        for i in range(nsubjects):
            self.cov = self.cov + (np.outer(param_est[i,:], param_est[i,:]) + hess_inv[:, :, i]) - np.outer(self.mu, self.mu)

        self.cov = self.cov/nsubjects
#
# TODO: ADD MLE, EMPIRICALPRIORS, MCMC, VB, AND GAUSSIAN PROCESSES
#
#class MLE(object):
#    def __init__(self, loglik_func, params):
#        self.loglik_func = loglik_func
#        self.params = params
#        self.nparams = len(params)
#
#    def fit(self, data):
#        pass
#
#class EmpiricalPriors(object):
#    def __init__(self):
#        pass
#
#class MCMC(object):
#    def __init__(self):
#        pass
#
#class VB(object):
#    def __init__(self):
#        pass
#
#class GP(object):
#    def __init__(self):
#        pass

#===============================================================================
#
#   FITRFIT
#       Object storing the optimization data
#
#===============================================================================

class fitrfit(object):
    """
    Class representing the results of a fitrmodel optimization.

    Attributes
    ----------
    name : str
        Model identifier. We suggest using free-parameters as identifiers
    method : str
        Method employed in optimization.
    nsubjects : int
        Number of subjects fitted.
    nparams : int
        Number of free parameters in the fitted model.
    params : ndarray(shape=(nsubjects, nparams))
        Array of parameter estimates
    errs : ndarray(shape=(nsubjects, nparams))
        Array of parameter estimate errors
    nlogpost : ndarray(shape=(nsubjects))
        Subject level negative log-posterior probability
    nloglik : float
        Subject level negative log-likelihood
    LME : float
        Log-model evidence
    BIC : ndarray(shape=(nsubjects))
        Subject-level Bayesian Information Criterion
    AIC : ndarray(shape=(nsubjects))
        Subject-level Aikake Information Criterion

    Methods
    -------
    transform :
        Transforms data to constrained or unconstrained space
    """
    def __init__(self, method, nsubjects, nparams, name=None):
        self.name = name
        self.method = method
        self.nsubjects = nsubjects
        self.nparams = nparams
        self.params = np.zeros([nsubjects, nparams])
        self.paramnames = []
        self.hess = np.zeros([nparams, nparams, nsubjects])
        self.hess_inv = np.zeros([nparams, nparams, nsubjects])
        self.errs = np.zeros([nsubjects, nparams])
        self.nlogpost = np.zeros(nsubjects) + 1e7
        self.nloglik = np.zeros(nsubjects)
        self.LME = np.zeros(nsubjects)
        self.BIC = np.zeros(nsubjects)
        self.AIC = np.zeros(nsubjects)
        self.ts_LME = []
        self.ts_BIC = []
        self.ts_AIC = []

    def set_paramnames(self, params):
        """
        Sets the names of the RL parameters to the fitrfit object

        Parameters
        ----------
        params : list
            List of parameters from the rlparams module
        """
        for i in range(len(params)):
            self.paramnames.append(params[i].name)

    def plot_ae(self, actual, show_figure=True, save_figure=False, filename='actual-estimate.pdf'):
        """
        Plots actual parameters (if provided) against estimates

        Parameters
        ----------
        actual : ndarray(shape=(nsubjects, nparams))
            Array of actual parameters from a simulation
        show_figure : bool
            Whether to show figure output
        save_figure : bool
            Whether to save the figure to disk
        filename : str
            The file name to be output
        """
        nparams = np.shape(self.params)[1]
        fig, ax = plt.subplots(1, nparams)
        for i in range(nparams):
            maxval = np.maximum(np.max(actual[:,i]), np.max(self.params[:,i]))
            minval = np.minimum(np.min(actual[:,i]), np.min(self.params[:,i]))
            ax[i].plot(np.linspace(minval, maxval, 100), np.linspace(minval, maxval, 100))
            ax[i].scatter(actual[:,i], self.params[:,i])
            ax[i].set_xlabel('Actual')
            ax[i].set_ylabel('Estimate')
            ax[i].set_ylim([minval, maxval])
            ax[i].set_xlim([minval, maxval])

        if save_figure is True:
            plt.savefig(filename)

        if show_figure is True:
            plt.show()

    def plot_fit_ts(self, show_figure=True, save_figure=False, filename='fit-stats.pdf'):
        """
        Plots the log-model-evidence, BIC, and AIC over optimization iterations

        Parameters
        ----------
        show_figure : bool
            Whether to show figure output
        save_figure : bool
            Whether to save the figure to disk
        filename : str
            The file name to be output
        """
        n_opt_steps = len(self.ts_LME)
        fig, ax = plt.subplots(1, 3, figsize=(15, 5))

        #LME
        ax[0].plot(np.arange(n_opt_steps), self.ts_LME, lw=1.5, c='k')
        ax[0].scatter(np.arange(n_opt_steps), self.ts_LME, c='k')
        ax[0].set_xlabel('Optimization step')
        ax[0].set_title('Log Model Evidence (LME)\n')
        ax[0].set_xlim([0, n_opt_steps])

        #BIC
        ax[1].plot(np.arange(n_opt_steps), self.ts_BIC, lw=1.5, c='k')
        ax[1].scatter(np.arange(n_opt_steps), self.ts_BIC, c='k')
        ax[1].set_xlabel('Optimization step')
        ax[1].set_title('Bayesian Information Criterion\n')
        ax[1].set_xlim([0, n_opt_steps])

        #AIC
        ax[2].plot(np.arange(n_opt_steps), self.ts_AIC, lw=1.5, c='k')
        ax[2].scatter(np.arange(n_opt_steps), self.ts_AIC, c='k')
        ax[2].set_xlabel('Optimization step')
        ax[2].set_title('Aikake Information Criterion\n')
        ax[2].set_xlim([0, n_opt_steps])

        if save_figure is True:
            plt.savefig(filename)

        if show_figure is True:
            plt.show()

    def param_hist(self, show_figure=True, save_figure=False, filename='param-hist.pdf'):
        """
        Plots histograms of the parameter estimates

        Parameters
        ----------
        show_figure : bool
            Whether to show figure output
        save_figure : bool
            Whether to save the figure to disk
        filename : str
            The file name to be output

        """
        nparams = np.shape(self.params)[1]

        fig, ax = plt.subplots(1, nparams)
        for i in range(0, nparams):
            n, bins, patches = ax[i].hist(self.params[:,i], normed=1)
            y = mvn.pdf(bins, np.mean(self.params[:,i]), np.std(self.params[:,1]))
            ax[i].plot(bins, y, 'r--', lw=1.5)
            ax[i].set_title(self.paramnames[i] + '\n')

        if save_figure is True:
            plt.savefig(filename)

        if show_figure is True:
            plt.show()



"""
Functions that are used across fitr modules
"""
import numpy as np

def softmax(x):
    xmax = np.max(x)
    return np.exp(x-xmax)/np.sum(np.exp(x-xmax))

def logsumexp(x):
    """
    Numerically stable logsumexp.
    """
    xmax = np.max(x)
    y = xmax + np.log(np.sum(np.exp(x-xmax)))
    return y

def trans_UC(values_U, rng):
    'Transform parameters from unconstrained to constrained space.'
    if rng[0] == 'all_unc':
        return values_U
    values_T = []
    for value, rng in zip(values_U, rng):
        if rng   == 'unit':  # Range: 0 - 1.
            if value < -16.:
                value = -16.
            values_T.append(1./(1. + np.exp(-value)))  # Don't allow values smaller than 1e-
        elif rng   == 'half':  # Range: 0 - 0.5
            if value < -16.:
                value = -16.
            values_T.append(0.5/(1. + np.exp(-value)))  # Don't allow values smaller than 1e-7
        elif rng == 'pos':  # Range: 0 - inf
            if value > 16.:
                value = 16.
            values_T.append(np.exp(value))  # Don't allow values bigger than ~ 1e7.
        elif rng == 'unc': # Range: - inf - inf.
            values_T.append(value)
    return np.array(values_T)

def BIC(loglik, nparams, nsteps):
    """
    Calculates Bayesian information criterion
    """
    return nparams*np.log(nsteps) - 2*loglik

def AIC(nparams, loglik):
    """
    Calculates Aikake information criterion
    """
    return 2*nparams - 2*loglik

def LME(logpost, nparams, hessian):
    """
    Calculates log-model-evidence (LME)
    """
    return logpost + (nparams/2)*np.log(2*np.pi)-np.log(np.linalg.det(hessian))/2



"""
Log-likelihood functions for each task.
"""

import numpy as np
from .utils import logsumexp

#-------------------------------------------------------------------------------
#
#   TWO-STEP TASK
#       Log-likelihood functions for various models of the two-step task
#
#-------------------------------------------------------------------------------

class twostep_ll(object):
    def __init__(self, ptrans=0.7, rewards=[1, 0]):
        self.ptrans = np.array([1-ptrans, ptrans])

    def lr_cr_w(self, params, states, actions, rewards):
        """
        Likelihood function for model containing parameters (A) learning rate, (B) choice randomness, and (C) MB weight parameter
        """
        # Initialize parameters
        lr = params[0]
        cr = params[1]
        w  = params[2]

        # Initialize log-likelihood
        loglik = 0

        # Initialize Q arrays
        Q   = np.zeros([3, 2])
        Qmb = np.zeros([3, 2])
        Qmf = np.zeros([3, 2])

        ntrials = np.shape(states)[0]
        for t in range(ntrials):
            s1 = int(states[t,0])
            s2 = int(states[t,1])
            a1 = int(actions[t,0])
            a2 = int(actions[t,1])
            r = rewards[t]

            loglik = loglik + cr*Q[s2,a2] - logsumexp(cr*Q[s2,:])
            loglik = loglik + cr*Q[s1,a1] - logsumexp(cr*Q[s1,:])

            # Update model-free values
            Qmf[s2, a2] = Qmf[s2, a2] + lr*(r - Qmf[s2, a2])
            Qmf[s1, a1] = Qmf[s1, a1] + lr*(Qmf[s2, a2]-Qmf[s1, a1])

            # Update model based values
            Qmb[0, a1] = self.ptrans[1]*np.max(Qmf[1,:]) + self.ptrans[0]*np.max(Qmf[2,:])
            Qmb[0, a2] = self.ptrans[1]*np.max(Qmf[2,:]) + self.ptrans[0]*np.max(Qmf[1,:])

            # Linear combination of MF and MB
            Q = w*Qmb + (1-w)*Qmf


        return loglik

    def lr_cr_et_w(self, params, states, actions, rewards):
        """
        Likelihood function for model containing parameters (A) learning rate, (B) choice randomness, (C) eligibility trace, and (D) MB weight parameter
        """
        # Initialize parameters
        lr = params[0]
        cr = params[1]
        et = params[2]
        w  = params[3]

        # Initialize log-likelihood
        loglik = 0

        # Initialize Q arrays
        Q   = np.zeros([3, 2])
        Qmb = np.zeros([3, 2])
        Qmf = np.zeros([3, 2])

        ntrials = np.shape(states)[0]
        for t in range(ntrials):
            s1 = int(states[t,0])
            s2 = int(states[t,1])
            a1 = int(actions[t,0])
            a2 = int(actions[t,1])
            r = rewards[t]

            loglik = loglik + cr*Q[s2,a2] - logsumexp(cr*Q[s2,:])
            loglik = loglik + cr*Q[s1,a1] - logsumexp(cr*Q[s1,:])

            # Update model-free values
            Qmf[s2, a2] = Qmf[s2, a2] + lr*(r - Qmf[s2, a2])
            Qmf[s1, a1] = Qmf[s1, a1] + lr*(Qmf[s2, a2]-Qmf[s1, a1])*et

            # Update model based values
            Qmb[0, a1] = self.ptrans[1]*np.max(Qmf[1,:]) + self.ptrans[0]*np.max(Qmf[2,:])
            Qmb[0, a2] = self.ptrans[1]*np.max(Qmf[2,:]) + self.ptrans[0]*np.max(Qmf[1,:])

            # Linear combination of MF and MB
            Q = w*Qmb + (1-w)*Qmf


        return loglik

    def dummy(self, params, states, actions, rewards):
        """
        Likelihood function without learning
        """

        # Initialize parameters
        cr = params[0]

        # Initialize log-likelihood
        loglik = 0

        # Initialize Q arrays
        Q   = np.zeros([3, 2])
        Qmb = np.zeros([3, 2])
        Qmf = np.zeros([3, 2])

        ntrials = np.shape(states)[0]
        for t in range(ntrials):
            s1 = int(states[t,0])
            s2 = int(states[t,1])
            a1 = int(actions[t,0])
            a2 = int(actions[t,1])

            loglik = loglik + cr*Q[s2,a2] - logsumexp(cr*Q[s2,:])
            loglik = loglik + cr*Q[s1,a1] - logsumexp(cr*Q[s1,:])

        return loglik
#-------------------------------------------------------------------------------
#
#   BANDIT
#       Log-likelihood functions for various models of a bandit task
#
#-------------------------------------------------------------------------------

class bandit_ll(object):
    def __init__(self, narms=2):
        self.narms=narms

    def lr_cr(self, params, states, actions, rewards):
        """
        Likelihood function containing parameters (A) learning rate, and (B) choice randomness
        """
        lr = params[0]
        cr = params[1]

        ntrials = len(actions)
        Q = np.zeros(self.narms)
        loglik = 0

        for t in range(0, ntrials):
            a = int(actions[t])
            r = rewards[t]
            loglik = loglik + cr*Q[a] - logsumexp(cr*Q)
            Q[a] = Q[a] + lr*(r - Q[a])

        return loglik

    def lr_cr_rs(self, params, states, actions, rewards):
        """
        Likelihood function containing parameters (A) learning rate, (B) choice randomness, and (C) reward sensitivity
        """
        lr = params[0]
        cr = params[1]
        rs = params[2]

        ntrials = len(actions)
        Q = np.zeros(self.narms)
        loglik = 0

        for t in range(0, ntrials):
            a = int(actions[t])
            r = rewards[t]
            loglik = loglik + cr*Q[a] - logsumexp(cr*Q)
            Q[a] = Q[a] + lr*(rs*r - Q[a])

        return loglik

    def lrp_lrn_cr(self, params, states, actions, rewards):
        """
        Likelihood function containing parameters (A) positive learning rate, (B) negative learning rate, and (C) choice randomness
        """
        lrp = params[0]
        lrn = params[1]
        cr  = params[2]

        ntrials = len(actions)
        Q = np.zeros(self.narms)
        loglik = 0

        for t in range(0, ntrials):
            a = int(actions[t])
            r = rewards[t]
            loglik = loglik + cr*Q[a] - logsumexp(cr*Q)

            rpe = r-Q[a]
            if rpe >=0:
                lr = lrp
            else:
                lr = lrn

            Q[a] = Q[a] + lr*rpe

        return loglik

    def lrp_lrn_cr_rs(params, states, actions, rewards):
        """
        Likelihood function containing parameters (A) positive learning rate, (B) negative learning rate, (c) choice randomness, and (D) reward sensitivity
        """
        lrp = params[0]
        lrn = params[1]
        cr  = params[2]
        rs  = params[3]

        ntrials = len(actions)
        Q = np.zeros(self.narms)
        loglik = 0

        for t in range(0, ntrials):
            a = int(actions[t])
            r = rewards[t]
            loglik = loglik + cr*Q[a] - logsumexp(cr*Q)

            rpe = rs*r-Q[a]
            if rpe >=0:
                lr = lrp
            else:
                lr = lrn

            Q[a] = Q[a] + lr*rpe

        return loglik

    def dummy(self, params, states, actions, rewards):
        """
        Likelihood function without any learning
        """
        cr  = params[0]

        ntrials = len(actions)
        Q = np.zeros(self.narms)
        loglik = 0

        for t in range(0, ntrials):
            a = int(actions[t])
            loglik = loglik + cr*Q[a] - logsumexp(cr*Q)

        return loglik



"""
Classes representing each task
"""
import numpy as np
import matplotlib.pyplot as plt

class SyntheticData(object):
    """
    Object representing synthetic data

    Methods
    -------
    cumreward_param_plot :
        Plots the cumulative reward against model parameters. Useful to determine the relationship between reward acquisition and model parameters for a given task.
    plot_cumreward :
        Plots the cumulative reward over time for each subject
    """
    def __init__(self):
        self.data = {}
        self.params = None

    def cumreward_param_plot(self, alpha=0.9):
        if self.params is not None:
            nsubjects = np.shape(self.params)[0]
            creward = np.zeros(nsubjects)
            for i in range(0, nsubjects):
                creward[i] = np.sum(self.data[i]['R'])

            nparams = np.shape(self.params)[1]
            fig, ax = plt.subplots(1, nparams, figsize=(15, 5))
            for i in range(0, nparams):
                ax[i].scatter(self.params[:,i], creward, c='k', alpha=alpha)

            plt.show()
        else:
            print('ERROR: There are no parameters assigned')
            return

    def plot_cumreward(self):
        """ Plots cumulative reward over time for each subject"""
        nsubjects = len(self.data)
        fig, ax = plt.subplots(1, 1)
        for i in range(nsubjects):
            nsteps = len(self.data[i]['R'])
            ax.plot(np.arange(nsteps), np.cumsum(self.data[i]['R']))

        ax.set_title('Cumulative Reward by Subject\n')
        ax.set_xlabel('Trial')
        ax.set_ylabel('Reward')
        plt.show()



#===============================================================================
#
#   SIMPLE BANDIT TASK
#
#===============================================================================

class bandit(object):
    """
    Simple one-step bandit task.

    Attributes
    ----------
    narms : int
        Number of arms
    rewards : ndarray(shape=(2))
        First entry is the reward, if gained, and the second entry is the magnitude of the loss
    rprob : {ndarray(shape=(narms)), 'stochastic'}
        Probabilty of reward for each arm of the task. One can either specify the probabilities for each arm or enter 'stochastic,' which will vary the reward probability by a gaussian random walk

    Methods
    -------
    simulate(nsubjects,ntrials)
        Runs the task on simulated subjects

    """
    def __init__(self, narms=2, rewards=[1, 0], rprob='stochastic', rprob_sd=0.025, rprob_bounds=[0.2, 0.8]):
        self.narms = narms
        self.rewards = rewards
        self.rprob_sd = rprob_sd

        if rprob == 'stochastic':
            self.rprob_bounds = rprob_bounds
        else:
            if len(rprob) != self.narms:
                print('Reward probability vector must have one entry per arm')
                return
            if np.any(rprob > 1) or np.any(rprob < 0):
                print('Reward probabilities must all lie between 0 and 1')
                return

            self.rprob=rprob

    def simulate(self, ntrials, params):

        nsubjects = np.shape(params)[0]

        # Initialize reward paths
        path_max = self.rprob_bounds[0]
        path_min = self.rprob_bounds[1]
        path_sd  = self.rprob_sd

        results = SyntheticData()
        results.params = params

        for i in range(0, nsubjects):
            paths = np.random.uniform(path_min, path_max, size=[ntrials+1, 2])

            # initialize subject-level value table
            Q  = np.zeros(2)
            lr = params[i,0]
            cr = params[i,1]

            results.data[i] = {'S': np.zeros(ntrials),
                               'A': np.zeros(ntrials),
                               'R': np.zeros(ntrials),
                               'RPE': np.zeros(ntrials)}

            for t in range(0, ntrials):
                a = action(cr*Q)
                r = reward(a, paths[t, :])

                # learn
                rpe = r - Q[a]
                Q[a] = Q[a] + lr*rpe

                # store values
                results.data[i]['S'][t] = 0
                results.data[i]['A'][t] = a
                results.data[i]['R'][t] = r
                results.data[i]['RPE'][t] = rpe

                # update reward probabilities
                rand_step = np.random.normal(0, path_sd, size=2)
                paths[t+1, :] = np.maximum(np.minimum(paths[t,:] + rand_step, path_max), path_min)

        return results

#===============================================================================
#
#   TWO-STEP TASK
#
#===============================================================================

class twostep(object):
    """
    Model of the two-step task (Daw et al. 2011).

    Attributes
    ----------
    ptrans : ndarray
        Probability of transitioning from state 0 to either state 1 or 2 depending on the choice made at the first step of the task.

    Methods
    -------
    simulate
        Generates synthetic data from the task.

    References
    ----------
    [1] Daw, N.D. et al. (2011) Model-based influences on humans’ choices and striatal prediction errors. Neuron 69, 1204–1215
    """
    def __init__(self, ptrans=0.7, rewards=[1, 0]):
        self.ptrans = np.array([1-ptrans, ptrans])

    def simulate(self, ntrials, params):

        nsubjects = np.shape(params)[0]

        results = SyntheticData()
        results.params = params

        #initialize reward paths
        path_max = 0.8
        path_min = 0.2
        path_sd  = 0.025


        for i in range(nsubjects):
            # Initialize paths within subjects
            paths = np.random.uniform(path_min, path_max, size=[ntrials+1, 4])

            lr = params[i,0]
            cr = params[i,1]
            w  = params[i,2]

            results.data[i] = {
                'S': np.zeros([ntrials, 2]),
                'A': np.zeros([ntrials, 2]),
                'R': np.zeros(ntrials)
            }

            Q   = np.zeros([3, 2])
            Qmb = np.zeros([3, 2])
            Qmf = np.zeros([3, 2])

            for t in range(ntrials):
                s1 = int(0)
                a1 = int(action(cr*Q[s1,:]))

                s2 = int(np.random.binomial(1, p=self.ptrans[a1]) + 1)
                a2 = int(action(cr*Q[s2,:]))

                rprob = paths[t,:]
                rprob = np.reshape(rprob, (2, 2))
                r = np.random.binomial(1, p=rprob[s2-1, a2])

                # Update model-free values
                Qmf[s2, a2] = Qmf[s2, a2] + lr*(r - Qmf[s2, a2])
                Qmf[s1, a1] = Qmf[s1, a1] + lr*(Qmf[s2, a2]-Qmf[s1, a1])

                # Update model based values
                Qmb[0, a1] = self.ptrans[1]*np.max(Qmf[1,:]) + self.ptrans[0]*np.max(Qmf[2,:])
                Qmb[0, a2] = self.ptrans[1]*np.max(Qmf[2,:]) + self.ptrans[0]*np.max(Qmf[1,:])

                # Linear combination of MF and MB
                Q = w*Qmb + (1-w)*Qmf

                # Store data
                results.data[i]['S'][t,0] = s1
                results.data[i]['S'][t,1] = s2
                results.data[i]['A'][t,0] = a1
                results.data[i]['A'][t,1] = a2
                results.data[i]['R'][t]   = r

                # Update reward probabilities
                rand_step = np.random.normal(0, path_sd, size=4)
                paths[t+1, :] = np.maximum(np.minimum(paths[t,:] + rand_step, path_max), path_min)

        return results



#===============================================================================
#
#   UTILITY FUNCTIONS
#
#===============================================================================

def action(x):
    p = np.exp(x)/np.sum(np.exp(x))
    return np.argmax(np.random.multinomial(1, pvals=p))

def reward(a, paths):
    return np.random.binomial(1, p=paths[a])



"""
Functions for model selection/comparison.
"""
import numpy as np
import matplotlib.pyplot as plt
plt.style.use('seaborn-colorblind')

from scipy.special import digamma as psi
from scipy.special import gammaln


class ModelSelectionResult(object):
    """
    Object containing results of model selection

    Attributes
    ----------
    modelnames : list
        List of strings labeling models
    xp : ndarray
        Exceedance probabilities for each model
    pxp : ndarray
        Protected exceedance probabilities for each model
    BIC : ndarray
        Bayesian information criterion measures for each model
    AIC : ndarray
        Aikake information criterion measures for each model

    """
    def __init__(self, method):
        self.modelnames = []

        if method=='BMS':
            self.xp = []
            self.pxp = []

        if method=='BIC':
            self.BIC = []

        if method=='AIC':
            self.AIC = []

        if method=='All':
            self.xp = []
            self.pxp = []
            self.BIC = []
            self.AIC = []

    def plot(self, statistic, show_figure=True, save_figure=False, filename='modelselection-plot.pdf', figsize=(10, 10)):
        if statistic=='pxp':
            bar_height = self.pxp
            plot_title = 'Protected Exceedance Probabilities'
            plot_ylabel = 'Probability'
        elif statistic=='xp':
            bar_height = self.xp
            plot_title = 'Exceedance Probabilities'
            plot_ylabel = 'Probability'
        elif statistic=='BIC':
            bar_height = self.BIC
            plot_title = 'Bayesian Information Criterion'
            plot_ylabel = 'BIC'
        elif statistic=='AIC':
            bar_height = self.AIC
            plot_title = 'Aikake Information Criterion'
            plot_ylabel = 'AIC'

        width = 0.7
        ind = np.arange(len(self.modelnames))

        fig, ax = plt.subplots(1, 1, figsize=figsize)
        ax.bar(ind, bar_height, width=width, align="center")
        ax.set_ylabel(plot_ylabel)
        ax.set_xticks(ind)
        ax.set_xticklabels(self.modelnames)
        ax.set_title(plot_title)

        if save_figure is True:
            plt.savefig(filename)

        if show_figure is True:
            plt.show()

class BIC(object):
    """
    Model comparison with Bayesian Information Criterion

    Attributes
    ----------
    modelfits : list
        List of fitrfit objects from completed model fitting
    """
    def __init__(self, model_fits):
        self.modelfits = model_fits

    def run(self):
        """
        Runs model comparison by Bayesian Information Criterion
        """
        results = ModelSelectionResult(method='BIC')

        for i in range(len(self.modelfits)):
            results.BIC.append(self.modelfits[i].name)

        return results

class AIC(object):
    """
    Model comparison with Aikake Information Criterion

    Attributes
    ----------
    modelfits : list
        List of fitrfit objects from completed model fitting
    """
    def __init__(self, model_fits):
        self.modelfits = model_fits

    def run(self):
        """
        Runs model comparison by Aikake Information Criterion
        """
        results = ModelSelectionResult(method='AIC')

        for i in range(len(self.modelfits)):
            results.AIC.append(self.modelfits[i].name)

        return results

class BMS(object):
    """
    Bayesian model selection

    Attributes
    ----------
    modelfits : list
        List of fitrfit objects from completed model fitting
    nmodels : int
        Number of models to be compared
    nsubjects: int
        Number of subjects in the sample. (Must be equal across all fits).
    c_limit : float
        Threshold at which to stop model comparison

    Methods
    -------
    run
        Runs the Bayesian model selection algorithm
    dirichlet_exceedance
        Computes exceedance probabilities for a Dirichlet distribution


    References
    ----------
    [1] Rigoux, L. et al. (2014) Neuroimage 84, 971–985
    [2] Samuel Gershman's _mfit_ package (on GitHub)
    """
    def __init__(self, model_fits, c_limit=10e-100):
        self.modelfits = model_fits

        # Extract number of models and number of subjects
        self.nmodels = len(self.modelfits)
        self.nsubjects = self.modelfits[0].nsubjects
        self.c_limit = c_limit

    def run(self):
        """
        Runs Bayesian model selection algorithm

        Returns
        -------
        ModelComparisonResult :
            Object representing model comparison results
        """

        bms_results = {}

        # Preallocate memory
        u = np.zeros([self.nsubjects, self.nmodels])
        g = np.zeros([self.nsubjects, self.nmodels])
        a = np.zeros([self.nsubjects, self.nmodels])
        lme = np.zeros([self.nsubjects, self.nmodels])

        # Initialize Dirichlet priors
        alpha0 = np.ones(self.nmodels)
        alpha = alpha0

        bms_iter = 0
        prediction_error = 1

        print('===== STARTING BAYESIAN MODEL SELECTION =====\n' +
              'Number of models: ' + str(self.nmodels) + '\n' +
              'Subjects: ' + str(self.nsubjects) + '\n' +
              'Convergence limit: ' + str(self.c_limit) + '\n' +
              '=============================================\n')

        while prediction_error > self.c_limit:
            bms_iter += 1
            print('[BMS] ITERATION: ' + str(bms_iter))

            for i in range(self.nsubjects):
                log_u = np.zeros(self.nmodels)
                for j in range(self.nmodels):
                    test_lme = self.modelfits[j].LME[i]

                    # If Hessian was degenerate, use BIC approximation of LME
                    lmefinite = np.isfinite(test_lme)
                    lmecomplex = np.iscomplex(test_lme)
                    if lmefinite is False or lmecomplex is True:
                        lme[i, j] = -0.5*self.modelfits[j].BIC[i]
                        - self.modelfits[j].nparams*np.log(2*np.pi)
                    else:
                        lme[i, j] = self.modelfits[j].LME[i]

                    log_u[j] = lme[i, j] + psi(alpha[j]) - psi(np.sum(alpha))

                u[i, :] = np.exp(log_u - np.max(log_u))
                g[i, :] = u[i, :]/np.sum(u[i, :])
                a[i, :] = np.random.multinomial(1, pvals=g[i, :])

            beta = np.sum(a, axis=0)
            alpha_previous = alpha
            alpha = alpha0 + beta

            # Test for convergence
            prediction_error = np.linalg.norm(alpha - alpha_previous)

        # Find expected values of model probabilities
        bms_results = {
                'pms': g,
                'r': alpha/np.sum(alpha),
                'xp': self.dirichlet_exceedance(alpha),
                'pxp': []
                }

        posterior = {
            'a': alpha,
            'r': g.T
        }
        priors = {'a': alpha0}
        bor = self.BOR(L=lme.T, posterior=posterior, priors=priors)
        bms_results['pxp'] = np.dot(bms_results['xp'], (1-bor))
        + bor/self.nmodels

        # Store data in ModelSelectionResult object
        results = ModelSelectionResult(method='BMS')
        for i in range(len(self.modelfits)):
            results.modelnames.append(self.modelfits[i].name)
        results.pxp = bms_results['pxp']
        results.xp = bms_results['xp']

        return results

    def dirichlet_exceedance(self, alpha):
        """
        Computes exceedance probabilities for a Dirichlet distribution

        Parameters
        ----------
        alpha :

        Returns
        -------
        xp :
            Exceedance probabilities

        References
        ----------
        [1] Samuel Gershman's _mfit_ package (on GitHub)
        """

        nsamples = 1e6
        K = len(alpha)

        # Perform sampling in blocks
        blk = np.ceil(nsamples*K*8/(2**28))
        blk = np.floor(nsamples/blk * np.ones(blk))
        blk[-1] = nsamples - np.sum(blk[:-1])

        xp = np.zeros(K)
        for i in range(0, len(blk)):
            # Sample from univariate gamma then normalize
            r = np.zeros([int(blk[i]), K])
            for k in range(0, K):
                r[:, k] = np.random.gamma(alpha[k], 1, size=int(blk[i]))

            sr = np.sum(r, axis=1)
            for k in range(K):
                r[:, k] = r[:, k]/sr

            # Compute Exceedance probabilities:
            #   For any given model k1, compute probability that
            #    it is more likely than any other model k2 != k1
            j = np.argmax(r, axis=1)
            xp = xp + np.histogram(j, bins=K)[0]

        xp = xp/nsamples
        return xp

    def BOR(self, L, posterior, priors, C=None):
        """
        Computes Bayes Omnibus Risk (BOR)

        Parameters
        ----------
        L :
        posterior:
        priors:
        C:

        Returns
        -------
        bor :
            Bayesian omnibus risk

        References
        ----------
        [1] Samuel Gershman's _mfit_ package (on GitHub)
        """

        if C is None:
            options = {'families': False}

            # Evidence of null (equal model frequencies)
            F0 = self.FE_null(L, options)[0]
        else:
            options = {
                'families': True,
                'C': C
            }

            # Evidence of null (equal model frequencies) under family prior
            F0 = self.FE_null(L, options)[1]

        # Evidence of alternative
        F1 = self.FE(L=L, posterior=posterior, priors=priors)

        bor = 1/(1+np.exp(F1-F0))
        return bor

    def FE(self, L, posterior, priors):
        """
        Derives free energy for current approximate posterior distribution

        Parameters
        ----------
        L :
        posterior :
        priors :

        Returns
        -------
        F :
            Free energy of the current posterior

        References
        ----------
        [1] Rigoux L., Daunizeau J. _VBA Toolbox_
        """
        nmodels, nsubjects = np.shape(L)
        a0 = np.sum(posterior['a'])
        Elogr = psi(posterior['a']) - psi(np.sum(posterior['a']))
        Sqf = np.sum(gammaln(posterior['a']))
        - gammaln(a0) - np.sum((posterior['a']-1)*Elogr)
        Sqm = 0

        for i in range(nsubjects):
            Sqm = Sqm - np.sum(posterior['r'][:, i] *
                               np.log(posterior['r'][:, i]+np.spacing(1)))

        gln_sumprior = gammaln(np.sum(priors['a']))
        sum_glnprior = np.sum(gammaln(priors['a']))
        ELJ = gln_sumprior - sum_glnprior + np.sum((priors['a']-1)*Elogr)

        for i in range(nsubjects):
            for k in range(nmodels):
                ELJ = ELJ + posterior['r'][k, i]*(Elogr[k]+L[k, i])

        F = ELJ + Sqf + Sqm

        return F

    def FE_null(self, L, options):
        """
        Derives the free energy of the 'null' hypothesis

        Parameters
        ----------
        L :
        options :

        Returns
        -------
        F0m :
            Evidence for the null (i.e. equal probabilities) over models
        F0f :
            Evidence for the null (i.e. equal probabilities) over families

        References
        ----------
        [1] Rigoux L., Daunizeau J. _VBA Toolbox_
        """
        nmodels, nsubjects = np.shape(L)
        if options['families'] is True:
            f0 = np.dot(options['C'],
                        np.sum(options['C'],
                        axis=0))**(-1/(np.shape(options['C'])[1]))
            F0f = 0
        else:
            F0f = []

        F0m = 0
        for i in range(nsubjects):
            tmp = L[:, i] - np.max(L[:, i])
            g = np.exp(tmp)/np.sum(np.exp(tmp))
            for k in range(nmodels):
                lme_err = L[k, i]-np.log(nmodels)-np.log(g[k] + np.spacing(1))
                F0m = F0m + g[k]*lme_err
                if options['families'] is True:
                    logf0 = np.log(f0[k])
                    logGeps = np.log(g[k]+np.spacing(1))
                    F0f = F0f + g[k] * (L[k, i]-logGeps+logf0)

        return F0m, F0f



"""
Objects representing each parameter object
"""
import scipy.stats

class Param(object):
    def __init__(self, name=None, rng=None):
        self.name  = name
        self.range = rng
        self.dist  = None

class LearningRate(Param):
    """
    A learning rate object.
    """
    def __init__(self, name='Learning Rate', rng='unit'):
        self.name = name
        self.rng  = rng
        self.dist = scipy.stats.beta(1.1, 1.1)

class RewardSensitivity(Param):
    """
    A reward sensitivity object.
    """
    def __init__(self, name='Reward Sensitivity', rng='unit'):
        self.name = name
        self.rng  = rng
        self.dist = scipy.stats.beta(1.1, 1.1)

class EligibilityTrace(Param):
    """
    An eligibility trace parameter object.
    """
    def __init__(self, name='Eligibility Trace', rng='unit'):
        self.name = name
        self.rng  = rng
        self.dist = scipy.stats.beta(1.1, 1.1)

class MBMF_Balance(Param):
    """
    An object representing the parameter that balances model-based and model-free control.
    """
    def __init__(self, name='Model-Based Control Weight', rng='unit'):
        self.name = name
        self.rng  = rng
        self.dist = scipy.stats.beta(1.1, 1.1)

class ChoiceRandomness(Param):
    """
    An choice randomness parameter object
    """
    def __init__(self, name='Choice Randomness', rng='pos'):
        self.name = name
        self.rng  = rng
        self.dist = scipy.stats.gamma(5, 1)



from setuptools import setup, find_packages
from codecs import open

import os

# Get path to directory this script is run from
here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
# Open an close file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='fitr',
      version='0.0.0.dev1',
      author='Abraham Nunes',
      author_email='nunes@dal.ca',
      description='Fit reinforcement learning models to behavioural data',
      long_description=long_description,
      url='https://github.com/ComputationalPsychiatry/fitr',
      keywords=['reinforcement learning',
                'computational psychiatry',
                'python',
                'model fitting'],
      classifiers=[
        # Maturity of project
        #    2 - Pre-Alpha
        #    3 - Alpha
        #    4 - Beta
        #    5 - Production/Stable
        #    6 - Mature
        #    7 - Inactive
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 2.7'
      ],
      license='MIT',
      packages=find_packages(exclude=['docs', 'tests']),
      # include_package_data=True,
      # zip_safe=False
      )
